@@ -10,9 +10,11 @@
 # You are free to use and modify this code in anyway you find useful. Please leave this comment in the code
 # to acknowledge its original source. If you feel like it, I enjoy hearing about projects that use my code,
 # but don't feel like you have to let me know or ask permission.
+from __future__ import division
 
 import math
 import random
+import re
 
 TNUMBER = 0
 TOP1 = 1
@@ -27,7 +29,7 @@ class Token():
         self.type_ = type_
         self.index_ = index_ or 0
         self.prio_ = prio_ or 0
-        self.number_ = number_ or 0
+        self.number_ = number_ if number_ != None else 0
 
     def toString(self):
         if self.type_ == TNUMBER:
@@ -77,7 +79,7 @@ class Expression():
                     newexpression.append(nstack.pop(0))
                 newexpression.append(item)
         while nstack:
-            newexpression.add(nstack.pop(0))
+            newexpression.append(nstack.pop(0))
 
         return Expression(newexpression, self.ops1, self.ops2, self.functions)
 
@@ -109,8 +111,7 @@ class Expression():
         values = values or {}
         nstack = []
         L = len(self.tokens)
-        for i in range(0, L):
-            item = self.tokens[i]
+        for item in self.tokens:
             type_ = item.type_
             if type_ == TNUMBER:
                 nstack.append(item.number_)
@@ -133,11 +134,11 @@ class Expression():
             elif type_ == TFUNCALL:
                 n1 = nstack.pop()
                 f = nstack.pop()
-                if f.apply and f.call:
+                if callable(f):
                     if type(n1) is list:
-                        nstack.append(f.apply(None, n1))
+                        nstack.append(f(*n1))
                     else:
-                        nstack.append(f.call(None, n1))
+                        nstack.append(f(n1))
                 else:
                     raise Exception(f + ' is not a function')
             else:
@@ -153,7 +154,10 @@ class Expression():
             item = self.tokens[i]
             type_ = item.type_
             if type_ == TNUMBER:
-                nstack.append(item.number_)
+                if type(item.number_) == str:
+                    nstack.append("'"+item.number_+"'")
+                else:
+                    nstack.append( item.number_)
             elif type_ == TOP2:
                 n2 = nstack.pop()
                 n1 = nstack.pop()
@@ -161,20 +165,26 @@ class Expression():
                 if toJS and f == '^':
                     nstack.append('math.pow(' + n1 + ',' + n2 + ')')
                 else:
-                    nstack.append('({n1}{f}{n2})'.format(
+                    frm='({n1}{f}{n2})'
+                    if f == ',':
+                        frm = '{n1}{f}{n2}'
+
+                    nstack.append(frm.format(
                         n1=n1,
                         n2=n2,
                         f=f,
                     ))
+
+
             elif type_ == TVAR:
                 nstack.append(item.index_)
             elif type_ == TOP1:
                 n1 = nstack.pop()
                 f = item.index_
                 if f == '-':
-                    nstack.append('(' + f + n1 + ')')
+                    nstack.append('(' + f + str(n1) + ')')
                 else:
-                    nstack.append(f + '(' + n1 + ')')
+                    nstack.append(f + '(' + str(n1) + ')')
             elif type_ == TFUNCALL:
                 n1 = nstack.pop()
                 f = nstack.pop()
@@ -185,7 +195,10 @@ class Expression():
             raise Exception('invalid Expression (parity)')
         return nstack[0]
 
-    def variables(self):
+    def __str__(self):
+        return self.toString()
+
+    def symbols(self):
         vars = []
         for i in range(0, len(self.tokens)):
             item = self.tokens[i]
@@ -193,157 +206,16 @@ class Expression():
                 vars.append(item.index_)
         return vars
 
+    def variables(self):
+        return [
+            sym for sym in self.symbols()
+            if sym not in self.functions]
+
 
 class Parser:
 
-    class Expression():
-
-        def __init__(self, tokens, ops1, ops2, functions):
-            self.tokens = tokens
-            self.ops1 = ops1
-            self.ops2 = ops2
-            self.functions = functions
-
-        def simplify(self, values):
-            values = values or {}
-            nstack = []
-            newexpression = []
-            L = len(self.tokens)
-            for i in range(0, L):
-                item = self.tokens[i]
-                type_ = item.type_
-                if type_ == TNUMBER:
-                    nstack.append(item)
-                elif type_ == TVAR and item.index_ in values:
-                    item = Token(TNUMBER, 0, 0, values[item.index_])
-                    nstack.append(item)
-                elif type_ == TOP2 and len(nstack) > 1:
-                    n2 = nstack.pop()
-                    n1 = nstack.pop()
-                    f = self.ops2[item.index_]
-                    item = Token(TNUMBER, 0, 0, f(n1.number_, n2.number_))
-                    nstack.append(item)
-                elif type_ == TOP1 and nstack:
-                    n1 = nstack.pop()
-                    f = self.ops1[item.index_]
-                    item = Token(TNUMBER, 0, 0, f(n1.number_))
-                    nstack.append(item)
-                else:
-                    while len(nstack) > 0:
-                        newexpression.append(nstack.pop(0))
-                    newexpression.append(item)
-            while nstack:
-                newexpression.add(nstack.pop(0))
-
-            return Expression(newexpression, self.ops1, self.ops2, self.functions)
-
-        def substitute(self, variable, expr):
-            if not isinstance(expr, Expression):
-                pass #expr = Parser().parse(str(expr))
-            newexpression = []
-            L = len(self.tokens)
-            for i in range(0, L):
-                item = self.tokens[i]
-                type_ = item.type_
-                if type_ == TVAR and item.index_ == variable:
-                    for j in range(0, len(expr.tokens)):
-                        expritem = expr.tokens[j]
-                        replitem = Token(
-                            expritem.type_,
-                            expritem.index_,
-                            expritem.prio_,
-                            expritem.number_,
-                        )
-                        newexpression.append(replitem)
-                else:
-                    newexpression.append(item)
-
-            ret = Expression(newexpression, self.ops1, self.ops2, self.functions)
-            return ret
-
-        def evaluate(self, values):
-            values = values or {}
-            nstack = []
-            L = len(self.tokens)
-            for i in range(0, L):
-                item = self.tokens[i]
-                type_ = item.type_
-                if type_ == TNUMBER:
-                    nstack.append(item.number_)
-                elif type_ == TOP2:
-                    n2 = nstack.pop()
-                    n1 = nstack.pop()
-                    f = self.ops2[item.index_]
-                    nstack.append(f(n1, n2))
-                elif type_ == TVAR:
-                    if item.index_ in values:
-                        nstack.append(values[item.index_])
-                    elif item.index_ in self.functions:
-                        nstack.append(self.functions[item.index_])
-                    else:
-                        raise Exception('undefined variable: ' + item.index_)
-                elif type_ == TOP1:
-                    n1 = nstack.pop()
-                    f = self.ops1[item.index_]
-                    nstack.append(f(n1))
-                elif type_ == TFUNCALL:
-                    n1 = nstack.pop()
-                    f = nstack.pop()
-                    if f.apply and f.call:
-                        if type(n1) is list:
-                            nstack.append(f.apply(None, n1))
-                        else:
-                            nstack.append(f.call(None, n1))
-                    else:
-                        raise Exception(f + ' is not a function')
-                else:
-                    raise Exception('invalid Expression')
-            if len(nstack) > 1:
-                raise Exception('invalid Expression (parity)')
-            return nstack[0]
-
-        def toString(self, toJS=False):
-            nstack = []
-            L = len(self.tokens)
-            for i in range(0, L):
-                item = self.tokens[i]
-                type_ = item.type_
-                if type_ == TNUMBER:
-                    nstack.append(item.number_)
-                elif type_ == TOP2:
-                    n2 = nstack.pop()
-                    n1 = nstack.pop()
-                    f = item.index_
-                    if toJS and f == '^':
-                        nstack.append('math.pow(' + n1 + ',' + n2 + ')')
-                    else:
-                        nstack.append('(' + n1 + f + n2 + ')')
-                elif type_ == TVAR:
-                    nstack.append(item.index_)
-                elif type_ == TOP1:
-                    n1 = nstack.pop()
-                    f = item.index_
-                    if f == '-':
-                        nstack.append('(' + f + n1 + ')')
-                    else:
-                        nstack.append(f + '(' + n1 + ')')
-                elif type_ == TFUNCALL:
-                    n1 = nstack.pop()
-                    f = nstack.pop()
-                    nstack.append(f + '(' + n1 + ')')
-                else:
-                    raise Exception('invalid Expression')
-            if len(nstack) > 1:
-                raise Exception('invalid Expression (parity)')
-            return nstack[0]
-
-        def variables(self):
-            vars = []
-            for i in range(0, len(self.tokens)):
-                item = self.tokens[i]
-                if item.type_ == TVAR and not item.index_ in vars:
-                    vars.append(item.index_)
-            return vars
+    class Expression(Expression):
+        pass
 
     PRIMARY      = 1
     OPERATOR     = 2
@@ -370,14 +242,50 @@ class Parser:
     def mod(self, a, b):
         return a % b
 
-    def concat(self, a, b):
-        return u'{0}{1}'.format(a, b)
+    def concat(self, a, b,*args):
+        result=u'{0}{1}'.format(a, b)
+        for arg in args:
+            result=u'{0}{1}'.format(result, arg)
+        return result
+
+    def equal (self, a, b ):
+        return a == b
+
+    def notEqual (self, a, b ):
+        return a != b
+
+    def greaterThan (self, a, b ):
+        return a > b
+
+    def lessThan (self, a, b ):
+        return a < b
+
+    def greaterThanEqual (self, a, b ):
+        return a >= b
+
+    def lessThanEqual (self, a, b ):
+        return a <= b
+
+    def andOperator (self, a, b ):
+        return ( a and b )
+
+    def orOperator (self, a, b ):
+        return  ( a or  b )
+
+    def xorOperator (self, a, b ):
+        return  ( a ^ b )
+
+    def inOperator(self, a, b):
+        return a in b
+
+    def notOperator(self, a):
+        return not a
 
     def neg(self, a):
         return -a
 
     def random(self, a):
-        return math.random() * (a or 1)
+        return random.random() * (a or 1)
 
     def fac(self, a):  # a!
         return math.factorial(a)
@@ -385,6 +293,36 @@ class Parser:
     def pyt(self, a, b):
         return math.sqrt(a * a + b * b)
 
+    def sind(self, a):
+        return math.sin(math.radians(a))
+
+    def cosd(self, a):
+        return math.cos(math.radians(a))
+
+    def tand(self, a):
+        return math.tan(math.radians(a))
+
+    def asind(self, a):
+        return math.degrees(math.asin(a))
+
+    def acosd(self, a):
+        return math.degrees(math.acos(a))
+
+    def atand(self, a):
+        return math.degrees(math.atan(a))
+
+    def roll(self, a, b):
+        rolls = []
+        roll = 0
+        final = 0
+        for c in range(1, a):
+            roll = random.randint(1, b)
+            rolls.append(roll)
+        return rolls
+
+    def ifFunction(self,a,b,c):
+        return b if a else c
+
     def append(self, a, b):
         if type(a) != list:
             return [a, b]
@@ -410,13 +348,21 @@ class Parser:
             'asin': math.asin,
             'acos': math.acos,
             'atan': math.atan,
+
+            'sind': self.sind,
+            'cosd': self.cosd,
+            'tand': self.tand,
+            'asind': self.asind,
+            'acosd': self.acosd,
+            'atand': self.atand,
+
             'sqrt': math.sqrt,
-            'log': math.log,
             'abs': abs,
             'ceil': math.ceil,
             'floor': math.floor,
             'round': round,
             '-': self.neg,
+            'not': self.notOperator,
             'exp': math.exp,
         }
 
@@ -427,18 +373,33 @@ class Parser:
             '/': self.div,
             '%': self.mod,
             '^': math.pow,
+            '**': math.pow,
             ',': self.append,
             '||': self.concat,
+            "==": self.equal,
+            "!=": self.notEqual,
+            ">": self.greaterThan,
+            "<": self.lessThan,
+            ">=": self.greaterThanEqual,
+            "<=": self.lessThanEqual,
+            "and": self.andOperator,
+            "or": self.orOperator,
+            "xor": self.xorOperator,
+            "in": self.inOperator,
+            "D": self.roll
         }
 
         self.functions = {
-            'random': random,
+            'random': self.random,
             'fac': self.fac,
+            'log': math.log,
             'min': min,
             'max': max,
             'pyt': self.pyt,
             'pow': math.pow,
             'atan2': math.atan2,
+            'concat':self.concat,
+            'if': self.ifFunction
         }
 
         self.consts = {
@@ -486,12 +447,19 @@ class Parser:
             if self.isOperator():
                 if self.isSign() and expected & self.SIGN:
                     if self.isNegativeSign():
-                        self.tokenprio = 2
+                        self.tokenprio = 5
                         self.tokenindex = '-'
                         noperators += 1
                         self.addfunc(tokenstack, operstack, TOP1)
                     expected = \
                         self.PRIMARY | self.LPAREN | self.FUNCTION | self.SIGN
+                elif self.isLogicalNot() and expected & self.SIGN:
+                    self.tokenprio = 2
+                    self.tokenindex = 'not'
+                    noperators += 1
+                    self.addfunc(tokenstack, operstack, TOP1)
+                    expected = \
+                        self.PRIMARY | self.LPAREN | self.FUNCTION | self.SIGN
                 elif self.isComment():
                     pass
                 else:
@@ -512,7 +480,7 @@ class Parser:
                     self.error_parsing(self.pos, 'unexpected string')
                 token = Token(TNUMBER, 0, 0, self.tokennumber)
                 tokenstack.append(token)
-                expected = self.OPERATOR or self.RPAREN | self.COMMA
+                expected = self.OPERATOR | self.RPAREN | self.COMMA
             elif self.isLeftParenth():
                 if (expected & self.LPAREN) == 0:
                     self.error_parsing(self.pos, 'unexpected \"(\"')
@@ -588,7 +556,7 @@ class Parser:
 
     def error_parsing(self, column, msg):
         self.success = False
-        self.errormsg = 'parse error [column ' + str(column) + ']: ' + msg
+        self.errormsg = 'parse error [column ' + str(column) + ']: ' + msg + ', expression: ' + self.expression
         raise Exception(self.errormsg)
 
     def addfunc(self, tokenstack, operstack, type_):
@@ -607,13 +575,31 @@ class Parser:
 
     def isNumber(self):
         r = False
+
+        if self.expression[self.pos] == 'E':
+            return False
+
+        # number in scientific notation
+        pattern = r'([-+]?([0-9]*\.?[0-9]*)[eE][-+]?[0-9]+).*'
+        match = re.match(pattern, self.expression[self.pos: ])
+        if match:
+            self.pos += len(match.group(1))
+            self.tokennumber = float(match.group(1))
+            return True
+
+        # number in decimal
         str = ''
         while self.pos < len(self.expression):
             code = self.expression[self.pos]
             if (code >= '0' and code <= '9') or code == '.':
-                str += self.expression[self.pos]
+                if (len(str) == 0 and code == '.' ):
+                    str = '0'
+                str += code
                 self.pos += 1
-                self.tokennumber = float(str)
+                try:
+                    self.tokennumber = int(str)
+                except ValueError:
+                    self.tokennumber = float(str)
                 r = True
             else:
                 break
@@ -670,17 +656,18 @@ class Parser:
                 else:
                     buffer.append(c)
 
-        return buffer.join('')
+        return ''.join(buffer)
 
     def isString(self):
         r = False
         str = ''
         startpos = self.pos
-        if self.pos < len(self.expression) and self.expression[self.pos] == '':
+        if self.pos < len(self.expression) and self.expression[self.pos] in ("'", "\""):
+            quote_type = self.expression[self.pos]
             self.pos += 1
             while self.pos < len(self.expression):
                 code = self.expression[self.pos]
-                if code != '\'' or str[-1] == '\\':
+                if code != quote_type or (str != '' and str[-1] == '\\'):
                     str += self.expression[self.pos]
                     self.pos += 1
                 else:
@@ -695,42 +682,47 @@ class Parser:
             L = len(i)
             str = self.expression[self.pos:self.pos+L]
             if i == str:
-                self.tokennumber = self.consts[i]
-                self.pos += L
-                return True
+                if len(self.expression) <= self.pos + L:
+                    self.tokennumber = self.consts[i]
+                    self.pos += L
+                    return True
+                if not self.expression[self.pos + L].isalnum() and self.expression[self.pos + L] != "_":
+                    self.tokennumber = self.consts[i]
+                    self.pos += L
+                    return True
         return False
 
     def isOperator(self):
-        code = self.expression[self.pos]
-        if code == '+':
-            self.tokenprio = 0
-            self.tokenindex = '+'
-        elif code == '-':
-            self.tokenprio = 0
-            self.tokenindex = '-'
-        elif code == '|':
-            if self.expression[self.pos] == '|':
-                self.pos += 1
-                self.tokenprio = 0
-                self.tokenindex = '||'
-            else:
-                return False
-        elif code == '*':
-            self.tokenprio = 1
-            self.tokenindex = '*'
-        elif code == '/':
-            self.tokenprio = 2
-            self.tokenindex = '/'
-        elif code == '%':
-            self.tokenprio = 2
-            self.tokenindex = '%'
-        elif code == '^':
-            self.tokenprio = 3
-            self.tokenindex = '^'
-        else:
-            return False
-        self.pos += 1
-        return True
+        ops = (
+            ('**', 8, '**'),
+            ('^', 8, '^'),
+            ('%', 6, '%'),
+            ('/', 6, '/'),
+            (u'\u2219', 5, '*'), # bullet operator
+            (u'\u2022', 5, '*'), # black small circle
+            ('*', 5, '*'),
+            ('+', 4, '+'),
+            ('-', 4, '-'),
+            ('||', 3, '||'),
+            ('==', 3, '=='),
+            ('!=', 3, '!='),
+            ('<=', 3, '<='),
+            ('>=', 3, '>='),
+            ('<', 3, '<'),
+            ('>', 3, '>'),
+            ('in', 3, 'in'),
+            ('not ', 2, 'not'),
+            ('and ', 1, 'and'),
+            ('xor ', 0, 'xor'),
+            ('or ', 0, 'or'),
+        )
+        for token, priority, index in ops:
+            if self.expression.startswith(token, self.pos):
+                self.tokenprio = priority
+                self.tokenindex = index
+                self.pos += len(token)
+                return True
+        return False
 
     def isSign(self):
         code = self.expression[self.pos - 1]
@@ -744,6 +736,10 @@ class Parser:
         code = self.expression[self.pos - 1]
         return code == '-'
 
+    def isLogicalNot(self):
+        code = self.expression[self.pos - 4: self.pos]
+        return code == 'not '
+
     def isLeftParenth(self):
         code = self.expression[self.pos]
         if code == '(':
@@ -762,7 +758,12 @@ class Parser:
 
     def isComma(self):
         code = self.expression[self.pos]
-        return code == ','
+        if code==',':
+            self.pos+=1
+            self.tokenprio=-1
+            self.tokenindex=","
+            return True
+        return False
 
     def isWhite(self):
         code = self.expression[self.pos]
@@ -781,7 +782,7 @@ class Parser:
             str += c
         if len(str) > 0 and str in self.ops1:
             self.tokenindex = str
-            self.tokenprio = 5
+            self.tokenprio = 9
             self.pos += len(str)
             return True
         return False
@@ -796,22 +797,25 @@ class Parser:
             str += c
         if len(str) > 0 and (str in self.ops2):
             self.tokenindex = str
-            self.tokenprio = 5
+            self.tokenprio = 9
             self.pos += len(str)
             return True
         return False
 
     def isVar(self):
         str = ''
+        inQuotes = False
         for i in range(self.pos, len(self.expression)):
             c = self.expression[i]
             if c.lower() == c.upper():
-                if i == self.pos or (c != '_' and (c < '0' or c > '9')):
+                if ((i == self.pos and c != '"') or (not (c in '_."') and (c < '0' or c > '9'))) and not inQuotes :
                     break
+            if c == '"':
+                inQuotes = not inQuotes
             str += c
         if str:
             self.tokenindex = str
-            self.tokenprio = 4
+            self.tokenprio = 6
             self.pos += len(str)
             return True
         return False
